#include "global.h"
#include "AnimatedModel.h"
#include "Bone.h"
#include "AnimationSet.h"
#include "keyframes.h"
#include "GameException.h"

//DirectX has some jacked up header that it includes non-static, non-extern
//data, so only one cpp file can ever include it!
extern unsigned char D3DRM_XTEMPLATES[];
#define D3DRM_XTEMPLATE_BYTES 3278

const float UNITS_PER_SEC = 3600.0f;
const float FRAMES_PER_SEC = 24.0f;

/**
 * Struct for loading raw X file data.
 */
struct XFilePosKey
{
    DWORD t;
    DWORD count;
    Vector Pos;
};

/**
 * Struct for loading raw X file data.
 */
struct XFileScaleKey
{
    DWORD t;
    DWORD count;
    Vector Scale;
};

/**
 * Struct for loading raw X file data.
 */
struct XFileRotKey
{
    DWORD t;
    DWORD count;
    //NOTE x files are w x y z and QUATERNIONS are x y z w
    float w;
    float x;
    float y;
    float z;
};

using namespace std;

AnimatedModel::AnimatedModel( const std::string& modelName ) {
    HRESULT r = 0;
    LPDIRECTXFILE xFile = NULL;
    LPDIRECTXFILEENUMOBJECT xEnum = NULL;
    LPDIRECTXFILEDATA xObj = NULL;

    try {
        //File interface create and open with enum object.
        r = DirectXFileCreate( &xFile );
        if ( FAILED( r ) ) throw GameException( "DirectXFileCreate Failure", r );

        r = xFile->RegisterTemplates((LPVOID)D3DRM_XTEMPLATES, D3DRM_XTEMPLATE_BYTES);
        if ( FAILED( r ) ) throw GameException( "XFile RegisterTemplates Failure", r );

        string modelXFileName = modelName + ".x";
        r = xFile->CreateEnumObject( (LPVOID)modelXFileName.c_str(), DXFILELOAD_FROMFILE, &xEnum);
        if ( FAILED( r ) ) throw GameException( "XFile CreateEnumObject Failure", r );

        //Load bones/mesh
        r = xEnum->GetNextDataObject( &xObj );
        if ( FAILED( r ) ) throw GameException( "Error finding skeletal information", r );
        loadSkeleton( xObj );
        SAFERELEASE( xObj );

        //Load the animations
        r = xEnum->GetNextDataObject( &xObj );
        if ( FAILED( r ) ) throw GameException( "Error finding animation information", r );
        loadAnimations( xObj, modelName );
        SAFERELEASE( xObj );

        //Check to make sure file isn't more complicated than we can handle right now
        r = xEnum->GetNextDataObject( &xObj );
        if ( SUCCEEDED( r ) ) throw GameException( "This parser is limited right now and can only handle a single mesh on a single skeleton with a single animation set." );        

    } catch ( GameException& e ) {
        //Cleanup
        SAFERELEASE( xFile );
        SAFERELEASE( xEnum );
        SAFERELEASE( xObj );
        cleanUp();
        e.addMessage( "For AnimatedModel \"" + modelName + "\"" );
        throw e;
    }

    //Cleanup
    SAFERELEASE( xFile );
    SAFERELEASE( xEnum );
    SAFERELEASE( xObj );
}

void AnimatedModel::loadAnimations( LPDIRECTXFILEDATA xObj, const std::string& modelName ) {
    const GUID* oType;
    if ( FAILED( xObj->GetType( &oType ) ) || *oType != TID_D3DRMAnimationSet )
        throw GameException( "XFile Parser: The second root data in the file must be the AnimationSet." );

    LPDIRECTXFILEDATA xChildDat;
    LPDIRECTXFILEOBJECT xChildObj;
    AKeysList keysList;

    while ( SUCCEEDED( xObj->GetNextObject( &xChildObj ) ) ) {
        if ( SUCCEEDED( xChildObj->QueryInterface( IID_IDirectXFileData, (LPVOID*)&xChildDat ) ) ) {
            loadSingleAnim( keysList, xChildDat );
            xChildDat->Release();
        } else {
            xChildObj->Release();
            throw GameException( "XFile Parser: Unrecognized element in XFile." );
        }
        xChildObj->Release();
    }

    parseAnmFile( keysList, modelName );
}

void AnimatedModel::loadSingleAnim( AKeysList& keysList, LPDIRECTXFILEDATA xObj ) {
    const GUID* oType;
    if ( FAILED( xObj->GetType( &oType ) ) || *oType != TID_D3DRMAnimation )
        throw GameException( "XFile Parser: Unexpected element -- expected Animation." );

    HRESULT r = 0;
    LPDIRECTXFILEOBJECT xChild = NULL;
    LPDIRECTXFILEDATA xData = NULL;
    LPDIRECTXFILEDATAREFERENCE xRef = NULL;
    AnimationKeys keys;
    keys.boneName = "";

    while ( SUCCEEDED( xObj->GetNextObject( &xChild ) ) ) {
        r = xChild->QueryInterface( IID_IDirectXFileDataReference, (LPVOID*)&xRef );
        if ( SUCCEEDED( r ) ) {
            //We just got the reference to the bone (we hope).
            if ( keys.boneName != "" )
                throw GameException( "XFile Parser: More than one reference to a Bone was found." );

            //Resolve the reference
            r = xRef->Resolve( &xData );
            if ( FAILED( r ) )
                throw GameException( "XFile Parser: Error resolving Animation reference to Bone." );

            if ( FAILED( xData->GetType( &oType ) ) || *oType != TID_D3DRMFrame )
                throw GameException( "XFile Parser: Unexpected reference (expected ref to Frame)." );

            keys.boneName = getDXDataObjectName( xData );
            xData->Release();
            xRef->Release();

        } else {
            //Not a reference, so this should be animation data
            r = xChild->QueryInterface( IID_IDirectXFileData, (LPVOID*)&xData );

            if ( FAILED( xData->GetType( &oType ) ) || *oType != TID_D3DRMAnimationKey )
                throw GameException( "XFile Parser: Unexpected element -- expected AnimationKey." );

            void* dataPtr;
            DWORD dataSize;

            r = xData->GetData( NULL, &dataSize, &dataPtr );
            if ( FAILED( r ) )
                throw GameException( "XFile Parser: Error getting AnimationKey data." );

            DWORD keyType = ((DWORD*)dataPtr)[0];
            DWORD keyCount = ((DWORD*)dataPtr)[1];

            loadSingleKeySet( keys, dataPtr, keyType, keyCount );
        }
    }

    keysList.push_back( keys );
}

void AnimatedModel::loadSingleKeySet( AnimationKeys& keys, void* dataPtr, DWORD keyType, DWORD keyCount ) {
    switch ( keyType ) {
        case 0: { //rotation
            if ( !keys.rKeys.empty() )
                throw GameException( "XFile Parser: Multiple rotation key sets found." );

            XFileRotKey* rotKeys = (XFileRotKey*)((char*)dataPtr + (sizeof(DWORD) * 2) );
            RotKey rotKey;

            for ( DWORD i = 0; i < keyCount; ++i ) {
                rotKey.time = (float)rotKeys[i].t / UNITS_PER_SEC;
                rotKey.rot.w = rotKeys[i].w;
                rotKey.rot.x = -rotKeys[i].x;
                rotKey.rot.y = -rotKeys[i].y;
                rotKey.rot.z = -rotKeys[i].z;

                if ( !keys.rKeys.empty() )
                    if ( keys.rKeys[ keys.rKeys.size()-1 ].time >= rotKey.time )
                        throw GameException( "XFile Parser: Rotation keys are not ordered by time." );

                keys.rKeys.push_back( rotKey );
            }

            break; }

        case 1: {//scale
            if ( !keys.sKeys.empty() )
                throw GameException( "XFile Parser: Multiple scale key sets found." );

            XFileScaleKey* sKeys = (XFileScaleKey*)((char*)dataPtr + (sizeof(DWORD) * 2) );
            VectKey sKey;

            for ( DWORD i = 0; i < keyCount; ++i ) {
                sKey.time = (float)sKeys[i].t / UNITS_PER_SEC;
                sKey.v = sKeys[i].Scale;

                if ( !keys.sKeys.empty() )
                    if ( keys.sKeys[ keys.sKeys.size()-1 ].time >= sKey.time )
                        throw GameException( "XFile Parser: Scale keys are not ordered by time." );

                keys.sKeys.push_back( sKey );
            }

            break;}

        case 2: {//translation
            if ( !keys.tKeys.empty() )
                throw GameException( "XFile Parser: Multiple translation key sets found." );

            XFilePosKey* tKeys = (XFilePosKey*)((char*)dataPtr + (sizeof(DWORD) * 2) );
            VectKey tKey;

            for ( DWORD i = 0; i < keyCount; ++i ) {
                tKey.time = (float)tKeys[i].t / UNITS_PER_SEC;
                tKey.v = tKeys[i].Pos;

                if ( !keys.tKeys.empty() )
                    if ( keys.rKeys[ keys.tKeys.size()-1 ].time >= tKey.time )
                        throw GameException( "XFile Parser: Translation keys are not ordered by time." );

                keys.tKeys.push_back( tKey );
            }

            break;}

        case 3: //matrix
            throw GameException( "XFile Parser: Matrix key types are not supported." );
            break;

        default:
            throw GameException( "XFile Parser: Unknown key type " + toString( keyType ) + " found." );
            break;
    }
}

void AnimatedModel::parseAnmFile( AKeysList& keysList, const std::string& modelName ) {
    //We have all of the animations, so now we need to split them up into sets
    string animFileName = modelName + ".anm";
    ifstream animFile( animFileName.c_str() );
    string currLine;

    if ( !animFile )
        throw GameException( "Could not load " + animFileName + "." );

    enum { AS_STATE, BL_STATE, BL_REQ_STATE, EV_STATE, AN_STATE } currState = AS_STATE;
    int lineNo = 0;
    AnimationSet* currSet;
    BoneNameSet boneSet;

    while ( getline( animFile, currLine ) ) {
        ++lineNo;
        trimString( currLine );

        if ( currLine.length() > 0 && currLine[0] != '#' ) {
            string lineType;
            istringstream line( currLine );
            line >> lineType;

            try {
                if ( currState == AS_STATE ) {
                    if ( lineType != "as" )
                        throw GameException( "Invalid line type " + lineType + "; expected as." );

                    currSet = parseAsLine( line );
                    currState = BL_REQ_STATE;

                } else if ( currState == BL_REQ_STATE ) {
                    if ( lineType == "bl" ) {
                        parseBlLine( line, boneSet );
                        currState = BL_STATE;
                    } else {
                        throw GameException( "Invalid line type " + lineType + "; expected bl." );
                    }

                } else if ( currState == BL_STATE ) {
                    if ( lineType == "bl" ) {
                        parseBlLine( line, boneSet );

                    } else if ( lineType == "ev" ) {
                        //parseEvLine( line, eventList );
                        currState = EV_STATE;

                    } else if ( lineType == "an" ) {
                        parseAnLine( line, keysList, boneSet, *currSet );
                        currState = AN_STATE;

                    } else {
                        throw GameException( "Invalid line type " + lineType + "; expected bl, ev, or an." );
                    }

                } else if ( currState == EV_STATE ) {
                    if ( lineType == "ev" ) {
                        //parseEvLine( line, eventList );

                    } else if ( lineType == "an" ) {
                        parseAnLine( line, keysList, boneSet, *currSet );
                        currState = AN_STATE;

                    } else {
                        throw GameException( "Invalid line type " + lineType + "; expected ev or an." );
                    }

                } else if ( currState == AN_STATE ) {
                    if ( lineType == "an" ) {
                        parseAnLine( line, keysList, boneSet, *currSet );

                    } else if ( lineType == "as" ) {
                        currSet = parseAsLine( line );
                        boneSet.clear();
                        currState = BL_REQ_STATE;

                    } else {
                        throw GameException( "Invalid line type " + lineType + "; expected as or an." );
                    }
                }
            } catch ( GameException& e ) {
                e.addMessage( "On line " + toString( lineNo ) + " of file " + animFileName + "." );
                throw e;
            }
        }
    }

    if ( currState != AN_STATE )
        throw GameException( "Unfinished animation set on line " + toString( lineNo ) + " of file " + animFileName + "." );

    if ( aSets.empty() ) {
        GameException e( "Warning, no animation sets found in the file " + animFileName + "." );
        e.showMessageBox();
    }
}

AnimationSet* AnimatedModel::parseAsLine( std::istringstream& line ) {
    string asName;
    line >> asName;
    if ( !line )
        throw GameException( "Error reading as name." );

    AnimationSet* aSet = new AnimationSet( asName );
    aSets.push_back( aSet );
    return aSet;
}

void AnimatedModel::parseBlLine( std::istringstream& line, BoneNameSet& boneSet ) {
    string boneName, tok;
    line >> boneName;

    bool incParent(false), incChildren(false);

    if ( !line )
        throw GameException( "Error reading bone name." );

    while ( line >> tok ) {
        if ( tok == "+parent" && !incParent )
            incParent = true;
        else if ( tok == "+children" && !incChildren )
            incChildren = true;
        else
            throw GameException( "Invalid token " + tok + "." );
    }

    Bone* b = getSkeleton()->findBone( boneName );
    if ( !b )
        throw GameException( "Could not find the bone " + boneName );

    b->loadBoneNameSet( boneSet, incParent, incChildren );
}

void AnimatedModel::parseAnLine( std::istringstream& line, AKeysList& keysList, BoneNameSet& boneSet, AnimationSet& aSet ) {
    string animName;
    float startTime, endTime;

    line >> startTime >> endTime >> animName;
    if ( !line )
        throw GameException( "Error parsing an line." );

    startTime /= FRAMES_PER_SEC;
    endTime /= FRAMES_PER_SEC;

    aSet.addAnimation( keysList, boneSet, *getSkeleton(), startTime, endTime, animName );
}

void AnimatedModel::cleanUp() {
    ASetListIter iter = aSets.begin();
    for ( ; iter != aSets.end(); ++iter )
        delete (*iter);
    aSets.clear();

    SkinnedModel::cleanUp();
}

AnimatedModel::~AnimatedModel() {
    cleanUp();
}

int AnimatedModel::findSet( const std::string& setName ) const {
    for ( int c = 0; c < getNumSets(); ++c )
        if (aSets[c]->getName() == setName)
            return c;
    return -1;
}

int AnimatedModel::getNumSets() const {
    return (int)aSets.size();
}

const std::string& AnimatedModel::getSetName( int setIdx ) const {
    assert( 0 <= setIdx && setIdx < getNumSets() );
    return aSets[setIdx]->getName();
}

int AnimatedModel::findAnim( int setIdx, const std::string& animName ) const {
    assert( 0 <= setIdx && setIdx < getNumSets() );
    return aSets[setIdx]->findAnim( animName );
}

int AnimatedModel::getNumAnims( int setIdx ) const {
    assert( 0 <= setIdx && setIdx < getNumSets() );
    return aSets[setIdx]->getNumAnims();
}

const std::string& AnimatedModel::getAnimName( int setIdx, int animIdx ) const {
    assert( 0 <= setIdx && setIdx < getNumSets() );
    return aSets[setIdx]->getAnimName( animIdx );
}

int AnimatedModel::getActiveAnim( int setIdx ) const {
    assert( 0 <= setIdx && setIdx < getNumSets() );
    return aSets[setIdx]->getActiveAnim();
}

void AnimatedModel::setActiveAnim( int setIdx, int animIdx ) {
    assert( 0 <= setIdx && setIdx < getNumSets() );
    aSets[setIdx]->setActiveAnim( animIdx );
}

void AnimatedModel::setLooping( int setIdx, bool looping ) {
    assert( 0 <= setIdx && setIdx < getNumSets() );
    aSets[setIdx]->setLooping( looping );
}

bool AnimatedModel::isLooping( int setIdx ) const {
    assert( 0 <= setIdx && setIdx < getNumSets() );
    return aSets[setIdx]->isLooping();
}

void AnimatedModel::update( float dt ) {
    ASetListIter iter = aSets.begin();
    for ( ; iter != aSets.end(); ++iter )
        (*iter)->update( dt );
}

