#ifndef ANIMATEDMODEL_NFW342
#define ANIMATEDMODEL_NFW342

#include "global.h"
#include "SkinnedModel.h"
#include "keyframes.h"
class AnimationSet;

/**
 * The AnimatedModel class extends the functionality of the SkinnedModel class
 * to include animations on the skeleton. It is important to note that from
 * the perspective of the program, it is the skeleton that is being animated,
 * and not the skin or mesh. The SkinnedModel simple renders around the
 * skeleton. Thus the AnimatedModel focuses on tying sets of animation to
 * groups of joints in the Skeleton alone, and through the fact the skeleton
 * is moving, the functionality in the SkinnedModel will not need to "know"
 * about animation.
 * 
 * The animations for a model are split up into sets of animations. Each
 * AnimationSet cooresponds to a logical section of the model. For example if
 * a humanoid character was split at the waist, the two animation sets would
 * be "torso animations" and "leg animations." There is no limit on the number
 * of splits on a model (except that an animation set must contain at least
 * one joint). The number of active animations in an AnimatedModel is at most
 * equal to the number of sets it contains, since each AnimatedSet has a
 * currently active animation. In the previous example, the humanoid character
 * would have at most two active animations, one for the legs and one for the
 * torso, that both operate independently.
 */
class AnimatedModel : public SkinnedModel {
public:
    AnimatedModel( const std::string& modelName );

    /**
     * Destructor.
     */
    virtual ~AnimatedModel();

    /**
     * Searches through all of the AnimationSet objects for this model looking
     * for the set given by the specified name.  Returns an index to that set,
     * or -1 if the set is not found.
     */
    int findSet( const std::string& setName ) const;

    /**
     * Returns the number of animation sets for this model.
     */
    int getNumSets() const;

    /**
     * Returns the name for the animation set given by the index.
     *
     * @pre setIdx meets the requirements: 0 <= setIdx < getNumSets()
     */
    const std::string& getSetName( int setIdx ) const;

    /**
     * Finds the animation by the name animName in the specified animation
     * set.  Returns an index to the animation, or -1 if no animation by that
     * name is found.
     *
     * @pre setIdx meets the requirements: 0 <= setIdx < getNumSets()
     */
    int findAnim( int setIdx, const std::string& animName ) const;

    /**
     * Returns the number of animations for the given set.
     *
     * @pre setIdx meets the requirements: 0 <= setIdx < getNumSets()
     */
    int getNumAnims( int setIdx ) const;

    /**
     * Returns the name of the specified animation in the specified set.  As a
     * special case, the animIdx -1 will return the string "none", so that it
     * is easy to use this method with getActiveAnim.
     *
     * @pre setIdx meets the requirements: 0 <= setIdx < getNumSets()
     * @pre animIdx meets the requirements: -1 <= animIdx < getNumAnims(setIdx)
     */
    const std::string& getAnimName( int setIdx, int animIdx ) const;

    /**
     * Returns the index of the currently running animation for the given set,
     * or -1 if no animation is running.
     *
     * @see AnimationSet::getActiveAnim
     * @pre setIdx meets the requirements: 0 <= setIdx < getNumSets()
     */
    int getActiveAnim( int setIdx ) const;

    /**
     * Sets the currently active animation for this model.  Each set has at
     * most one active animation at a time.  If the passed animIdx is out of
     * range (negative numbers are always out of range), then that animation
     * set is disabled.
     * 
     * @pre setIdx meets the requirements: 0 <= setIdx < getNumSets()
     */
    void setActiveAnim( int setIdx, int animIdx );

    /**
     * Sets the looping flag of the specified set.
     *
     * @pre setIdx meets the requirements: 0 <= setIdx < getNumSets()
     *
     * @see AnimationSet::setLooping
     */
    void setLooping( int setIdx, bool looping );

    /**
     * Gets the looping flag of the specified set.
     *
     * @pre setIdx meets the requirements: 0 <= setIdx < getNumSets()
     *
     * @see AnimationSet::isLooping
     */
    bool isLooping( int setIdx ) const;

    /**
     * Updates the animation for this model.  Calling this method will result
     * in modifying the current position of the model's skeleton based on the
     * currently active animations.
     */
    void update( float dt );

protected:
    /**
     * Performs destructor functionality, used if exception is to be thrown
     * from contructor
     */
    virtual void cleanUp();

private:
    void loadAnimations( LPDIRECTXFILEDATA xObj, const std::string& modelName );
    void loadSingleAnim( AKeysList& keysList, LPDIRECTXFILEDATA xObj );
    void loadSingleKeySet( AnimationKeys& keysList, void* dataPtr, DWORD keyType, DWORD keyCount );

    void parseAnmFile( AKeysList& keysList, const std::string& modelName );
    AnimationSet* parseAsLine( std::istringstream& line );
    void parseBlLine( std::istringstream& line, BoneNameSet& boneSet );
    //void parseEvLine( std::istringstream& line, EventList& eventList );
    void parseAnLine( std::istringstream& line, AKeysList& keysList, BoneNameSet& boneSet, AnimationSet& aSet );

    typedef std::vector<AnimationSet*> ASetList;
    typedef ASetList::iterator ASetListIter;
    typedef ASetList::const_iterator ASetListCIter;

    ASetList aSets;
};

#endif

