#include "global.h"
#include "AnimationSet.h"
#include "SkeletalAnimation.h"

static const std::string& noneString( "none" );

AnimationSet::AnimationSet( const std::string& setName )
: name(setName), currTime( 0.0f ), looping( false ), activeAnim(NULL) {
}

AnimationSet::~AnimationSet() {
    AListIter iter = anims.begin();
    for ( ; iter != anims.end(); ++iter )
        delete (*iter);
}

void AnimationSet::addAnimation(
  const AKeysList& keys, const BoneNameSet& boneSet, Bone& rootBone,
  float startTime, float endTime, const std::string& animName ) {
    if ( findAnim( animName ) != -1 )
        throw GameException( "An animation by the name " + animName + " already exists in the set " + name + "." );
    try {
        anims.push_back( new SkeletalAnimation( keys, boneSet, rootBone, startTime, endTime, animName ) );
    } catch ( GameException& e ) {
        e.addMessage( "of AnimationSet " + name + "." );
        throw e;
    }
}

void AnimationSet::setLooping( bool looping ) {
    this->looping = looping;
}

bool AnimationSet::isLooping() const {
    return looping;
}

const std::string& AnimationSet::getName() const {
    return name;
}

int AnimationSet::getNumAnims() const {
    return (int)anims.size();
}

int AnimationSet::findAnim( const std::string& animName ) const {
    for ( int c = 0; c != getNumAnims(); ++c ) {
        if ( anims[c]->getName() == animName )
            return c;
    }
    return -1;
}

const std::string& AnimationSet::getAnimName( int anim ) const {
    if ( anim == -1 )
        return noneString;

    assert( anim >= 0 && anim < getNumAnims() );
    return anims[anim]->getName();
}

int AnimationSet::getActiveAnim() const {
    if ( activeAnim == NULL )
        return -1;

    for ( int idx = 0; idx < getNumAnims(); ++idx )
        if ( anims[idx] == activeAnim )
            return idx;

    assert( false );
    return -1;
}

void AnimationSet::setActiveAnim( int anim ) {
    if ( anim >= 0 && anim < getNumAnims() ) {
        currTime = 0.0f;
        activeAnim = anims[anim];
    } else
        activeAnim = NULL;
}

void AnimationSet::update( float dt ) {
    if ( activeAnim ) {
        //update time
        currTime += dt;

        //update for looping if needed.
        if ( looping ) {
            while ( currTime > activeAnim->getLength() )
                currTime -= activeAnim->getLength();
        } else {
            //Set animation to nothing if not looping.
            if ( currTime > activeAnim->getLength() ) {
                activeAnim = NULL;
                return;
            }
        }

        //perform animation
        activeAnim->animate( currTime );
    }
}
