#include "global.h"
#include "Bone.h"
#include "StaticModel.h"

using namespace std;

Bone::Bone() : name( "" ), parent(NULL) {
    D3DXMatrixIdentity( &refMat );
    D3DXMatrixIdentity( &offMat );
    D3DXMatrixIdentity( &transMat );
    D3DXMatrixIdentity( &finalMat );
}

Bone::Bone( const std::string& name, const Matrix& refMat,
            const Matrix& offMatt )
: name( name ), refMat( refMat ), offMat( offMat ),
  transMat( refMat ), finalMat( refMat ), parent( NULL ) {
}

Bone::~Bone() {
    BoneListIter iter = children.begin();
    for ( ; iter != children.end(); ++iter )
        delete (*iter);
}

void Bone::setName( std::string& name ) {
    this->name = name;
}

void Bone::setReferenceMatrix( const Matrix& m ) {
    refMat = m;
}

void Bone::setOffsetMatrix( const Matrix& m ) {
    offMat = m;
}
    
const std::string& Bone::getName() const {
    return name;
}

void Bone::loadBoneNameSet( BoneNameSet& boneSet, bool incParent, bool incChildren ) const {
    boneSet.insert( name );

    if ( incParent && parent )
        boneSet.insert( parent->getName() );

    if ( incChildren ) {
        BoneListCIter iter = children.begin();
        for ( ; iter != children.end(); ++iter )
            (*iter)->loadBoneNameSet( boneSet, false, true );
    }
}

Bone* Bone::findBone( const std::string& boneName ) {
    if ( name == boneName )
        return this;

    BoneListIter iter = children.begin();
    for ( ; iter != children.end(); ++iter ) {
        Bone* temp = (*iter)->findBone( boneName );
        if ( temp != NULL )
            return temp;
    }

    return NULL;
}

void Bone::setTransform( const Matrix& newMat ) {
    transMat = newMat;
}

void Bone::applyNewTransform( const Matrix& newMat ) {
    D3DXMatrixMultiply( &transMat, &newMat, &refMat );
}

void Bone::updateHierarchy( const Matrix& worldMat ) {
    D3DXMatrixMultiply( &finalMat, &transMat, &worldMat );

    BoneListIter iter = children.begin();
    for ( ; iter != children.end(); ++iter )
        (*iter)->updateHierarchy( finalMat );
}

const Matrix& Bone::getFinalMatrix() const {
    return finalMat;
}

const Matrix& Bone::getReferenceMatrix() const {
    return refMat;
}

void Bone::loadSkinMatrix( Matrix& m ) const {
    D3DXMatrixMultiply( &m, &offMat, &finalMat );
}

Bone* Bone::getParent() {
    return parent;
}

BoneListIter Bone::beginChildren() {
    return children.begin();
}

BoneListIter Bone::endChildren() {
    return children.end();
}

void Bone::addChild( Bone* newChild ) {
    assert( newChild->parent == NULL );
    newChild->parent = this;
    children.push_back( newChild );
}

void Bone::debugRender( StaticModel& m ) {
    GraphicsDevice d = g_dx->getDeviceHandle();
    HRESULT r = 0;

    r = d->SetTransform( D3DTS_WORLD, &finalMat );
    assert( SUCCEEDED( r ) );

    m.render();

    BoneListIter iter = children.begin();
    for ( ; iter != children.end(); ++iter )
        (*iter)->debugRender( m );
}

void Bone::printHierarchy( ostream& o, int level ) {
    o << string( level, ' ' ) << name << endl;

    BoneListIter iter = children.begin();
    for ( ; iter != children.end(); ++iter )
        (*iter)->printHierarchy( o, level+2 );
}
