#include "global.h"
#include "BoneInterpolator.h"
#include "Bone.h"
#include "Keyframes.h"

using namespace std;

/**
 * Utility method to select a group of keys based on an inclusive time range.
 */
template <class Collection>
void pullOutKeys( const Collection& src, Collection& dest, float startTime, float endTime ) {
    Collection::const_iterator iter = src.begin();
    for ( ; iter != src.end(); ++iter )
        if ( (*iter) >= startTime && (*iter) <= endTime )
            dest.push_back( *iter );
}

/**
 * Private utility class to serve as functor in a for_each call.
 */
class TimeShift {
public:
    explicit TimeShift( float shiftVal ) : shiftVal( shiftVal ) {}

    void operator () ( Key& key ) {
        key.time -= shiftVal;
    }

private:
    TimeShift();

    float shiftVal;
};

BoneInterpolator::BoneInterpolator( Bone* targetBone, const AnimationKeys& keys,
                                    float startTime, float endTime )
: target(targetBone) {
    assert( targetBone->getName() == keys.boneName );

    //pull out only the keys for our time range
    pullOutKeys( keys.sKeys, sKeys, startTime, endTime );
    pullOutKeys( keys.rKeys, rKeys, startTime, endTime );
    pullOutKeys( keys.tKeys, tKeys, startTime, endTime );

    RotKey k1 = rKeys[0];
    RotKey k2 = rKeys[rKeys.size()-1];
}

BoneInterpolator::~BoneInterpolator() {
}

bool BoneInterpolator::hasAnyKeys() const {
    return ( !sKeys.empty() || !rKeys.empty() || !tKeys.empty() );
}

float BoneInterpolator::getStartTime() {
    std::vector< float > temp;
    if ( !sKeys.empty() )
        temp.push_back( sKeys[0].time );
    if ( !rKeys.empty() )
        temp.push_back( rKeys[0].time );
    if ( !tKeys.empty() )
        temp.push_back( tKeys[0].time );

    assert( !temp.empty() );
    return *min_element( temp.begin(), temp.end() );
}

float BoneInterpolator::getEndTime() {
    std::vector< float > temp;
    if ( !sKeys.empty() )
        temp.push_back( sKeys[sKeys.size()-1].time );
    if ( !rKeys.empty() )
        temp.push_back( rKeys[rKeys.size()-1].time );
    if ( !tKeys.empty() )
        temp.push_back( tKeys[tKeys.size()-1].time );

    assert( !temp.empty() );
    return *max_element( temp.begin(), temp.end() );
}

void BoneInterpolator::timeShift( float shift ) {
    for_each( sKeys.begin(), sKeys.end(), TimeShift( shift ) );
    for_each( rKeys.begin(), rKeys.end(), TimeShift( shift ) );
    for_each( tKeys.begin(), tKeys.end(), TimeShift( shift ) );
}

/**
 * Given a time and a collection of keys, interpolates.  For times t outside
 * of the range of the animation, returns the first or last key, as
 * appropriate.
 *
 * @pre c contains at least 1 element
 */
template <class Collection>
Collection::value_type interpKeys( const Collection& c, float t ) {
    assert( !c.empty() );

    //Find first element >= t
    Collection::const_iterator iter = lower_bound( c.begin(), c.end(), t );
    if ( iter == c.end() )
        return *(c.end()-1);
    else if ( (*iter).time == t )
        return (*iter); //if t is exact, return key
    else {
        //do interpolation.  First find the location index.
        Collection::size_type loc = (Collection::size_type)(iter - c.begin());
        if ( loc <= 0 )
            return c[0];
        else
            return c[loc-1].interp( c[loc], t );
    }
}

void BoneInterpolator::animate( float t ) {
    Matrix result, temp;
    D3DXMatrixIdentity( &result );

    //Do scale, then rotate, then translate.  First 2 are optional.

    if ( !sKeys.empty() ) {
        VectKey key = interpKeys( sKeys, t );

        D3DXMatrixScaling( &result, key.v.x, key.v.y, key.v.z );
    }

    if ( !rKeys.empty() ) {
        RotKey key = interpKeys( rKeys, t );

        D3DXMatrixRotationQuaternion( &temp, &key.rot );
        D3DXMatrixMultiply( &result, &result, &temp );
    }

    if ( !tKeys.empty() ) {
        VectKey key = interpKeys( tKeys, t );

        D3DXMatrixTranslation( &temp, key.v.x, key.v.y, key.v.z );
        D3DXMatrixMultiply( &result, &result, &temp );
    } else {
        //use the position from the reference matrix if no trans keys
        const Matrix& ref = target->getReferenceMatrix();

        D3DXMatrixTranslation( &temp, ref._41, ref._42, ref._43 );
        D3DXMatrixMultiply( &result, &result, &temp );
    }

    target->setTransform( result );
}
