#ifndef BONEINTERPOLATOR_H_CIR341
#define BONEINTERPOLATOR_H_CIR341

#include "global.h"
#include "Keyframes.h"
class Bone;

/**
 * The BoneInterpolator does the real work of joint keyframing. Each
 * BoneInterpolator object references a single Bone (or joint), and all of the
 * keyframes for a single animation. The functionality of this class is
 * limited to simply calculating the position of a single bone given a time
 * value.
 */
class BoneInterpolator {
public:
    /**
     * Constructor for the BoneInterpolator.  The given keeys must be sorted
     * in ascending time order.  The given startTime and endTime are both
     * inclusive times to pull out the keys by.
     */
    BoneInterpolator( Bone* targetBone, const AnimationKeys& keys,
                      float startTime, float endTime );

    /**
     * Destructor.
     */
    ~BoneInterpolator();

    /**
     * Returns true if this BoneInterpolator has any keys.  If there are no
     * keys, then some methods are invalid to call.
     */
    bool hasAnyKeys() const;

    /**
     * Returns the minimum of time values across all keyframes (the start of
     * "interesting" data).
     *
     * @pre hasAnyKeys() returns true for ths object
     */
    float getStartTime();

    /**
     * Returns the maximum of time values across all keyframes (the end of
     * "interesting" data).
     *
     * @pre hasAnyKeys() returns true for ths object
     */
    float getEndTime();

    /**
     * Shifts the time values of all keyframes by adding the passed time value
     * to all keyframes.
     */
    void timeShift( float shift );

    /**
     * Animates the joint at the given time t.  If there is "looping" it
     * should be implemented by the caller.  If t is greater than the end time
     * for this animation (or less than start time), the joint position will
     * be constant, as the first or last key will be used without
     * interpolation.
     *
     * If there are no keys, then the translation from the reference matrix is
     * the result of the animation for all t.
     */
    void animate( float t );
    
private:
    BoneInterpolator();
    BoneInterpolator( const BoneInterpolator& o );
    BoneInterpolator& operator = ( const BoneInterpolator& rhs );
    
    RKeyList rKeys;
    SKeyList sKeys;
    TKeyList tKeys;

    Bone* target;
};

#endif

