/**
 * CG3 Skeletal Animation Project
 * Jason Winnebeck
 * May 12, 2003
 */

#include "global.h"
#include "GameDXListener.h"
#include "GameException.h"

using namespace std;

const int MAX_LIGHTS = 8;

//constructor
GameDX::GameDX() {
    //initialize all member pointers to zero
    m_pd3dFont = 0;
    m_pd3dDevice = 0;
    m_pd3d = 0;

    lights = new D3DLIGHT8[ MAX_LIGHTS ];
    lightEnabled = new bool[ MAX_LIGHTS ];
    for ( int c = 0; c < MAX_LIGHTS; ++c )
        lightEnabled[ c ] = false;

    D3DXMatrixIdentity( &worldMat );
    D3DXMatrixIdentity( &viewMat );
    D3DXMatrixIdentity( &projMat );

    defaultMat.Diffuse.a = 1.0f;
    defaultMat.Diffuse.b = defaultMat.Diffuse.g = defaultMat.Diffuse.r = 0.0f;
    defaultMat.Ambient.a = 1.0f;
    defaultMat.Ambient.b = defaultMat.Ambient.g = defaultMat.Ambient.r = 1.0f;
    defaultMat.Power = 10.0f;
    defaultMat.Emissive.a = defaultMat.Emissive.r = defaultMat.Emissive.g = defaultMat.Emissive.b = 0.1f;
    defaultMat.Specular.a = defaultMat.Specular.r = defaultMat.Specular.g = defaultMat.Specular.b = 1.0f;

    clearColor = D3DCOLOR_XRGB( 0, 0, 0 );
}

GameDX::~GameDX() {
    //release everything and return all pointers to zero
    releaseD3D();

    delete[] lights;
    delete[] lightEnabled;

    assert( listeners.empty() );
}

void GameDX::releaseD3D() {
    if (m_pd3dFont) {
        TRACE("Killing D3D Font");
        m_pd3dFont->Release();
        m_pd3dFont = 0;
    }
    if (m_pd3dDevice) {
        TRACE("Killing D3D Device");
        m_pd3dDevice->Release();
        m_pd3dDevice = 0;
    }
    if (m_pd3d ) {
        TRACE("Killing D3D COM Object");
        m_pd3d->Release();
        m_pd3d = 0;
    }
}

void GameDX::resetD3D() {
    HRESULT r = 0;

    r = setupPresentParams( (g_win->getWindowed() == TRUE) );
    if ( FAILED(r) )
        throw GameException( "Error setting up present parameters.", r );

    m_pd3dFont->Release();
    generateLostEvent();

    r = m_pd3dDevice->Reset( &d3dpp );
    if ( FAILED(r) ) {
        throw GameException( "Error resetting device.", r );
    }

    setupState();
    generateResetEvent();
}

void GameDX::Clear() {
    HRESULT r = 0;

    r = m_pd3dDevice->Clear( 0, 0, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,
                             clearColor, 1.0f, 0);

    if ( FAILED( r ) )
        throw GameException( "Error clearing device." );
}

void GameDX::SetClearColor( D3DCOLOR color ) {
    clearColor = color;

    HRESULT r = 0;
    r = m_pd3dDevice->SetRenderState( D3DRS_FOGCOLOR, clearColor );

    if ( FAILED( r ) )
        throw GameException( "Error setting fog color." );
}

void GameDX::SetTransform( D3DTRANSFORMSTATETYPE state, const D3DXMATRIX* pMatrix ) {
    switch ( state ) {
        case D3DTS_WORLD:
            worldMat = *pMatrix;
            break;
        case D3DTS_VIEW:
            viewMat = *pMatrix;
            break;
        case D3DTS_PROJECTION:
            projMat = *pMatrix;
            break;
        default:
            assert( false );
            throw GameException( "Invalid transform state change type." );
    };

    if ( FAILED( m_pd3dDevice->SetTransform( state, pMatrix ) ) )
        throw GameException( "Error trying to set transformation matrix." );
}

void GameDX::SetLight( DWORD index, const D3DLIGHT8* pLight ) {
    assert( index >= 0 && index < MAX_LIGHTS );

    lights[ index ] = *pLight;

    if ( FAILED( m_pd3dDevice->SetLight( index, pLight ) ) )
        throw GameException( "Error trying to set a light." );
}

void GameDX::LightEnable( DWORD LightIndex, BOOL bEnable ) {
    assert( LightIndex >= 0 && LightIndex < MAX_LIGHTS );

    lightEnabled[ LightIndex ] = ( bEnable == TRUE );

    if ( FAILED( m_pd3dDevice->LightEnable( LightIndex, bEnable ) ) )
        throw GameException( "Error trying to enable or disable a light." );
}

void GameDX::SetDefaultMaterial( const D3DMATERIAL8* newMat ) {
    defaultMat = *newMat;
}

void GameDX::RestoreDefaultMaterial() {
    if ( FAILED( m_pd3dDevice->SetMaterial( &defaultMat ) ) )
        throw GameException( "Error setting material." );
}

void GameDX::setupD3D() {
    HRESULT r = 0;

    m_pd3d = Direct3DCreate8( D3D_SDK_VERSION );
    if(m_pd3d == 0)
        throw GameException("GameDX: Could not create Direct3D Interface");

    r = setupPresentParams( (g_win->getWindowed() == TRUE) );
    if ( FAILED(r) )
        throw GameException( "Error setting up present parameters." );

    //Get device caps
    r = m_pd3d->GetDeviceCaps( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, &devCaps );
    if ( FAILED(r) )
        throw GameException( "Error getting device capabilities." );

    bool softReq = (devCaps.MaxVertexBlendMatrices < 4);
    if ( softReq ) {
        GameException w( "Warning: card does not support at least 4 blending matrices,\n" +
                         string( "reverting to software skin method.\n") +
                         "Your card supports " + toString( devCaps.MaxVertexBlendMatrices ) +
                         " blending matrices." );
        w.showMessageBox();
    }

    //get a handle to the device (watch out if you have two...)
    if ( !softReq ) {
        r = m_pd3d->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, g_win->getHwnd(), 
            D3DCREATE_HARDWARE_VERTEXPROCESSING, &d3dpp, &m_pd3dDevice);

        if( FAILED(r) )
            r = m_pd3d->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, g_win->getHwnd(), 
                                     D3DCREATE_MIXED_VERTEXPROCESSING, &d3dpp, &m_pd3dDevice);
    }

    if( softReq || FAILED(r) )
        r = m_pd3d->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, g_win->getHwnd(), 
            D3DCREATE_SOFTWARE_VERTEXPROCESSING, &d3dpp, &m_pd3dDevice);

    if( FAILED(r) )
        throw GameException( "GameDX: Could not create Direct3D Device", r );

    setupState();
}

LPDIRECT3DDEVICE8 & GameDX::getDeviceHandle() {
    return m_pd3dDevice;
}

LPDIRECT3D8 & GameDX::getD3D() {
    return m_pd3d;
}

D3DRECT & GameDX::getClientSurfaceCoords() {
    static D3DRECT rect = {0, 0, g_win->getWidth(), g_win->getHeight()};
    return rect;
}

LPD3DXFONT & GameDX::getFont() {
    return m_pd3dFont;
}

const D3DCAPS8& GameDX::getCaps() const {
    return devCaps;
}

using namespace std;

void GameDX::registerListener( GameDXListener& listener ) {
    LVectorIter iter = find( listeners.begin(), listeners.end(), &listener );
    if ( iter == listeners.end() )
        listeners.push_back( &listener );
}

void GameDX::removeListener( GameDXListener& listener ) {
    LVectorIter iter = find( listeners.begin(), listeners.end(), &listener );
    if ( iter != listeners.end() )
        listeners.erase( iter );
}

void GameDX::setupState() {
    HRESULT r = 0;

    //Get a font to draw with as neccessary
    r = D3DXCreateFont(
        m_pd3dDevice,
        (HFONT)GetStockObject(DEVICE_DEFAULT_FONT),
        &m_pd3dFont);
    if( FAILED(r) )
        throw GameException( "Error Creating Font" );

    //Setup Viewport
    D3DVIEWPORT8 vp;
    vp.X = 0;
    vp.Y = 0;
    vp.Width = g_win->getWidth();
    vp.Height = g_win->getHeight();
    vp.MinZ = 0.0f;
    vp.MaxZ = 1.0f;

    if ( FAILED( m_pd3dDevice->SetViewport( &vp ) ) )
        throw GameException( "Error setting viewport." );

    //clear the back buffer and then present
    m_pd3dDevice->BeginScene();
    m_pd3dDevice->Clear( 0, 0, D3DCLEAR_TARGET, D3DCOLOR_XRGB(0, 0, 0), 0.0f, 0);
    r = m_pd3dDevice->SetTextureStageState( 0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR );
    assert( SUCCEEDED( r ) );

    m_pd3dDevice->EndScene();
    m_pd3dDevice->Present(NULL, NULL, NULL, NULL);    

    //Setup matrices
    r = m_pd3dDevice->SetTransform( D3DTS_WORLD, &worldMat );
    if ( FAILED( r ) )
        throw GameException( "Error setting world matrix." );

    r = m_pd3dDevice->SetTransform( D3DTS_VIEW, &viewMat );
    if ( FAILED( r ) )
        throw GameException( "Error setting view matrix." );

    r = m_pd3dDevice->SetTransform( D3DTS_PROJECTION, &projMat );
    if ( FAILED( r ) )
        throw GameException( "Error setting projection matrix." );

    if ( FAILED( m_pd3dDevice->SetRenderState( D3DRS_LIGHTING, TRUE ) ) )
        throw GameException( "Error enabling lighting." );

    if ( FAILED( m_pd3dDevice->SetRenderState( D3DRS_SPECULARENABLE, TRUE ) ) )
        throw GameException( "Error enabling specular lighting." );

    //Setup lights
    for ( int c = 0; c < MAX_LIGHTS; ++c ) {
        if ( lightEnabled[ c ] ) {
            if ( FAILED( m_pd3dDevice->SetLight( (DWORD)c, &lights[ c ] ) ) ||
                 FAILED( m_pd3dDevice->LightEnable( (DWORD)c, TRUE ) ) )
                throw GameException( "Error setting light." );

        }
    }

    //Setup materials
    RestoreDefaultMaterial();
}

void GameDX::generateLostEvent() {
    LVectorIter iter = listeners.begin();
    for ( ; iter != listeners.end(); ++iter ) {
        (*iter)->OnLostDevice();
    }
}

void GameDX::generateResetEvent() {
    LVectorIter iter = listeners.begin();
    for ( ; iter != listeners.end(); ++iter ) {
        (*iter)->OnResetDevice();
    }
}

HRESULT GameDX::setupPresentParams( bool isWindowed ) {
    D3DDISPLAYMODE d3ddm;
    HRESULT r = m_pd3d->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &d3ddm);
    if( FAILED(r) ) {
        TRACE("Could Not Get Display Adapter Information");
        return r;
    }

    //create display parameters for device "creation"
    D3DFORMAT FullScreenFormat = D3DFMT_A8R8G8B8;

    ZeroMemory( &d3dpp, sizeof(D3DPRESENT_PARAMETERS) );
    d3dpp.Windowed = isWindowed;
    //tell DirectX that it is allowed to mess with the back buffer
    d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
    //set the back buffer to the same format as the primary surface
    d3dpp.BackBufferFormat = isWindowed ? d3ddm.Format : FullScreenFormat;
    d3dpp.BackBufferWidth = g_win->getWidth();
    d3dpp.BackBufferHeight = g_win->getHeight();
    d3dpp.BackBufferCount = 1;
    //set target window to render to
    d3dpp.hDeviceWindow = g_win->getHwnd();
    //turn on the depth buffer
    d3dpp.EnableAutoDepthStencil = TRUE;
    d3dpp.AutoDepthStencilFormat = D3DFMT_D16; //16bit depth buffer
    //set the refresh rate - not used if windowed
    d3dpp.FullScreen_RefreshRateInHz = isWindowed ? 0 : 60;
    //set the presentation interval - MUST be 0 if windowed
    // Use interval one to setup vsynch in full screen mode
    d3dpp.FullScreen_PresentationInterval = isWindowed ? 0 : D3DPRESENT_INTERVAL_ONE;

    return S_OK;
}
