/**
 * CG3 Skeletal Animation Project
 * Jason Winnebeck
 * May 12, 2003
 */

#ifndef _GAMEDX_H_
#define _GAMEDX_H_

#include "global.h"

class Window;
class GameDXListener;

/**
 * A container that holds and initializes all the DirectX objects
 */
class GameDX {

public:
    /**
     * Constructor
     *
     */
    GameDX();

    /**
     * Destructor
     *
     */
    ~GameDX();


public:
    /**
     * Sets up Direct3D
     *
     * @returns     the result
     * @exception GameException on failure
     */
    void setupD3D();

    /**
     * Gets a handle to the Direct3D device
     *
     * @returns     a handle to the Direct3D device
     */    
    LPDIRECT3DDEVICE8 & getDeviceHandle();

    /**
     * Gets a reference to the Direct3D object
     *
     * @returns     a reference to the Direct3D object
     */
    LPDIRECT3D8 & getD3D();

    /**
     * Gets a reference to the client surface coordinates
     *
     * @returns     a reference to the client surface coordinates
     */
    D3DRECT & getClientSurfaceCoords();

    /**
     * Gets a reference to a DirectX Font object
     *
     * @returns     a reference to a DirectX Font object
     */
    LPD3DXFONT & getFont();

    /**
     * Returns the capabilities of the current device.
     */
    const D3DCAPS8& getCaps() const;

    /**
     * Releases all objects created for Direct3D
     *
     */
    void releaseD3D();

    /**
     * After setting up the D3D device, this resets the device, picking up the
     * windowed mode from the global window.  This works essentially the same
     * as setupD3D, except that it allows re-initialization after a setupD3D
     * call.
     *
     * D3DPOOL_MANAGED textures and objects will remain between resets, but
     * not between a release/setup call pair.
     *
     * @exception GameException on failure
     */
    void resetD3D();

    /**
     * Performs a clear of the D3D surface and z-buffer.
     *
     * @exception GameException on failure
     */
    void Clear();

    /**
     * Sets the clear color to be used in Clear.  This also sets the fog color.
     *
     * @exception GameException on failure
     */
    void SetClearColor( D3DCOLOR color );

    /**
     * Sets the transform on this device.  Using this method ensures that the
     * matrix will be saved between device lost/reset transitions.  For view
     * and projection matrices, you should use this method.  For world
     * matricies that change only mid-frame, set the matrix directly with the
     * device, and set only the starting world matrix with this method.
     *
     * @exception GameException on failure
     */
    void SetTransform( D3DTRANSFORMSTATETYPE state, const D3DXMATRIX* pMatrix );

    /**
     * Sets the light, saving state in case of device transitions.
     *
     * @exception GameException on failure
     */
    void SetLight( DWORD index, const D3DLIGHT8* pLight );

    /**
     * Enables or disables the light, saving state in case of device transitions.
     *
     * @exception GameException on failure
     */
    void LightEnable( DWORD LightIndex, BOOL bEnable );

    /**
     * Sets the default material, but does not set it on the device.
     */
    void SetDefaultMaterial( const D3DMATERIAL8* );

    /**
     * Restores the default material by setting's the device's current material
     * to the default material.
     *
     * @exception GameException on failure
     */
    void RestoreDefaultMaterial();

    /**
     * Ensures that the listener given is registered.  If the given listener
     * has already registered, this method has no effect.
     */
    void registerListener( GameDXListener& listener );

    /**
     * Removes the given listener from the list of registered listeners, so
     * that it will no longer receive events.  If the given listener is not
     * in the list, this method will have no effect.
     */
    void removeListener( GameDXListener& listener );

private:

    /**
     * Sets up the common state after startup or a reset.
     *
     * @exception GameException on failure
     */
    void setupState();

    /**
     * Sends lost event to all listeners.
     */
    void generateLostEvent();

    /**
     * Sends reset event to all listeners.
     */
    void generateResetEvent();

    /**
     * Initializes the d3dpp variable.  This method requires the m_pd3d object
     * to be properly set up.
     */
    HRESULT setupPresentParams( bool isWindowed );

private:

    /**
     * Direct3D display mode
     *
     */
    D3DDISPLAYMODE d3ddm;

    /**
     * Direct3D present parameters
     *
     */
    D3DPRESENT_PARAMETERS d3dpp;

    /**
     * Direct3D object
     *
     */
    LPDIRECT3D8 m_pd3d;

    /**
     * Direct3D device
     *
     */
    LPDIRECT3DDEVICE8 m_pd3dDevice;

    /**
     * Direct3D Font object
     *
     */
    LPD3DXFONT m_pd3dFont;

    typedef std::vector< GameDXListener* > LVector; //Listener Vector
    typedef LVector::iterator LVectorIter;          //Listener Vector iterator

    LVector listeners;

    D3DXMATRIX worldMat;
    D3DXMATRIX viewMat;
    D3DXMATRIX projMat;

    D3DMATERIAL8 defaultMat;

    D3DLIGHT8* lights;
    bool* lightEnabled;

    D3DCOLOR clearColor;

    D3DCAPS8 devCaps;
};

#endif
