/**
 * CG3 Skeletal Animation Project
 * Jason Winnebeck
 * May 12, 2003
 */

#include "global.h"
#include "GameException.h"
#include "GameDXInput.h"

GameDXInput::GameDXInput()
{
    m_pDI = 0;
    m_pDIKeyboard = 0;
    m_pMouseDev = 0;
    m_pCursorSurf = 0;

    initialized = false;

    SetCursorPos( 0, 0 );

    for( int x = 0; x < KEYBUFFERSIZE; x++ )
    {
        m_KeyBuffer[x] = 0;
    }
    m_bKeyPollSuccess = false;
    m_bMousePollSuccess = false;
}

GameDXInput::~GameDXInput() 
{
    DITerm();
}

void GameDXInput::DITerm()
{
    //release everything and return all pointers to zero
    if (m_pDI)
    {
	    m_pDI->Release();
	    m_pDI = 0;
    }
    if (m_pDIKeyboard)
    {
	    m_pDIKeyboard->Release();
	    m_pDIKeyboard = 0;
    }
    if (m_pMouseDev)
    {
	    m_pMouseDev->Release();
	    m_pMouseDev = 0;
    }
    if (m_pCursorSurf)
    {
	    m_pCursorSurf->Release();
	    m_pCursorSurf = 0;
    }
}

void GameDXInput::SetupDI(HINSTANCE hInstance, HWND hwndMain)
{
    //find out about current display
    HRESULT r = 0;

    //*** Create the device ***
    r = DirectInput8Create(hInstance, DIRECTINPUT_VERSION, 
        IID_IDirectInput8, (void**)&m_pDI, NULL); 
    if ( FAILED( r ) )
        throw GameException( "Error GameDXInput.SetupDI DirectInput8Create" );

    //*** Get the keyboard ***
    // Release the device if it has already been created
    if( m_pDIKeyboard ) {
        m_pDIKeyboard->Unacquire();
        m_pDIKeyboard->Release();
    }

    r = m_pDI->CreateDevice(GUID_SysKeyboard, &m_pDIKeyboard, NULL); 
    if ( FAILED( r ) )
        throw GameException( "Error GameDXInput.SetupDI CreateDevice keyboard" );

    r = m_pDIKeyboard->SetDataFormat(&c_dfDIKeyboard); 
    if ( FAILED( r ) )
        throw GameException( "Error GameDXInput.SetupDI SetDataFormat keyboard" );

    // Set the cooperative level 
    r = m_pDIKeyboard->SetCooperativeLevel(hwndMain, 
                    DISCL_FOREGROUND | DISCL_NONEXCLUSIVE);

    if ( FAILED( r ) )
        throw GameException( "Error GameDXInput.SetupDI SetCooperativeLevel keyboard" );

    if (m_pDIKeyboard) m_pDIKeyboard->Acquire();

    //*** Get the mouse ***
    // Release the mouse device if it has already been created
    if( m_pMouseDev )
	    m_pMouseDev->Release();

    // Create the mouse device
    r = m_pDI->CreateDevice( GUID_SysMouse, &m_pMouseDev, NULL );
    if ( FAILED( r ) )
        throw GameException( "Error GameDXInput.SetupDI CreateDevice mouse" );

    // Set the cooperative level for the mouse
    r = m_pMouseDev->SetCooperativeLevel( hwndMain, DISCL_FOREGROUND | DISCL_NONEXCLUSIVE );
    if ( FAILED( r ) )
        throw GameException( "Error GameDXInput.SetupDI SetCooperativeLevel mouse" );

    // Set the data format for the mouse
    r = m_pMouseDev->SetDataFormat( &c_dfDIMouse );
    if ( FAILED( r ) )
        throw GameException( "Error GameDXInput.SetupDI SetDataFormat mouse" );

    initialized = true;
}

void GameDXInput::PollKeyboard()
{
    HRESULT r; 

    // Return if the object has not been initialized
    if( !initialized )
	    m_bKeyPollSuccess = false;

    //Reqcquire the deivce in case it was lost to another application
    r = m_pDIKeyboard->Acquire();
    while( r == DIERR_INPUTLOST ) 
            r = m_pDIKeyboard->Acquire();

    r = m_pDIKeyboard->GetDeviceState( sizeof(m_KeyBuffer),(LPVOID)&m_KeyBuffer ); 
    if FAILED(r) 
    { 
        m_bKeyPollSuccess = false;
        // If it failed, the device has probably been lost.
        if( r == DIERR_INPUTLOST ) { 
            while( r == DIERR_INPUTLOST ) 
                r = m_pDIKeyboard->Acquire();

            // Try to test the state again
            if( SUCCEEDED( r ) ) {
                 r = m_pDIKeyboard->GetDeviceState( sizeof(m_KeyBuffer),(LPVOID)&m_KeyBuffer );
                if( SUCCEEDED( r ) ) {
                    m_bKeyPollSuccess = true;
                }
            }
        }
    }
    // success!
    else {
        m_bKeyPollSuccess = true;
    }
}

bool GameDXInput::IsKeyDown( int keyToCheck )
{
    bool returnVal = false;  //Zero is do nothing
    
    if( m_bKeyPollSuccess && KEYPRESS( m_KeyBuffer, keyToCheck ) )
    {
        returnVal = true;
    }

    //Return the code of the key press, if any
    return returnVal;
}

void GameDXInput::PollMouse()
{
    HRESULT r;

    // Return if the object has not been initialized
    if( !initialized )
        m_bMousePollSuccess = false;

    // Reacquire the deivce in case it was lost to another application
    r = m_pMouseDev->Acquire();
    while( r == DIERR_INPUTLOST ) 
            r = m_pMouseDev->Acquire();

    // Get the state of the mouse
    r = m_pMouseDev->GetDeviceState( sizeof( DIMOUSESTATE ), &m_MouseData );
    if( FAILED( r ) )
    {
        m_bMousePollSuccess = false;
        // if failed, the device has problably been lossed
        if( r == DIERR_INPUTLOST )
        {
            // Reacquire the mouse
            while( r == DIERR_INPUTLOST )
                r = m_pMouseDev->Acquire();
            
            // Try to test the state again
            if( SUCCEEDED( r ) ) {
                r = m_pMouseDev->GetDeviceState( sizeof( DIMOUSESTATE ), &m_MouseData );
                if( SUCCEEDED( r ) ) {
                    m_bMousePollSuccess = true;
                }
            }
        }
    }
    // success!
    else {
        m_bMousePollSuccess = true;
    }
}

bool GameDXInput::IsButtonDown( int Button )
{
    bool retVal = false;
    
    // Return the button status from the buffer	
    if( m_bMousePollSuccess && KEYPRESS(m_MouseData.rgbButtons, Button ) ) {
	    retVal = true;
    }

    return retVal;
}

POINT GameDXInput::GetScreenPt()
{
    POINT ptScreen;
    GetCursorPos( &ptScreen );
    return ptScreen;
}

POINT GameDXInput::GetRelativePt()
{
    POINT relPt;
    relPt.x = m_MouseData.lX;
    relPt.y = m_MouseData.lY;
    return relPt;
}

POINT GameDXInput::GetWindowPt() {
    POINT ret( GetScreenPt() );
    ScreenToClient( g_win->getHwnd(), &ret );
    return ret;
}

void GameDXInput::ResetPos()
{
    SetCursorPos(0, 0);
}

void GameDXInput::CenterPos()
{
    POINT ret;
    ret.x = (LONG)(g_win->getWidth()/2.0f);
    ret.y = (LONG)(g_win->getHeight()/2.0f);

    ScreenToClient( g_win->getHwnd(), &ret );

    SetCursorPos( ret.x,ret.y );
}

void GameDXInput::CreateMouseSurf( LPDIRECT3DDEVICE8 pd3dDevice, char* pathName ) {
    //Create Surface for Mouse Pointer -----------------------------------
    HRESULT r = pd3dDevice->CreateImageSurface( 64, 64, D3DFMT_A8R8G8B8, &m_pCursorSurf );
    if( FAILED( r ) )
        throw new GameException( "GameDXPInput.CreateMouseSurf CreateImageSurface" );

    r = D3DXLoadSurfaceFromFile( m_pCursorSurf, 0, 0, pathName,
	    0, D3DX_FILTER_NONE, 0, 0);
    if( FAILED( r ) )
        throw new GameException( "GameDXPInput.CreateMouseSurf D3DXLoadSurfaceFromFile" );

    r = pd3dDevice->SetCursorProperties(0, 0, m_pCursorSurf );
    if( FAILED( r ) )
        throw new GameException( "GameDXPInput.CreateMouseSurf SetCursorProperties" );
}

void GameDXInput::ShowCursor( bool bShow, LPDIRECT3DDEVICE8 device ) {
    device->ShowCursor( bShow );
}
