/**
 * CG3 Skeletal Animation Project
 * Jason Winnebeck
 * May 12, 2003
 */

#ifndef GAMEDXINPUT_H
#define GAMEDXINPUT_H

#include "global.h"

#define KEYBUFFERSIZE 256
#define KEYPRESS(name, key) (name[key] & 0x80)

/**
 * GameDXInput
 * A wrapper class for Direct Input.  Handles creating and managing
 * a DirectInput object.  Both keyboard and mouse input are done through
 * this object.
 */
class GameDXInput {
public:
    /**
     * Constructor sets all member variables to 0
     */
    GameDXInput();

    /**
     * Destructor calls DITerm
     */
    ~GameDXInput();
    
    /**
     *  DITerm
     *
     *  Cleans up all the DirectInput pointers
     */
    void DITerm();

    /**
     *  SetupDI
     *
     *  Setups a DirectInput Object and ties it to the keyboard
     *
     * @param hInstance the HINSTANCE
     * @param hwndMain the main HWND
     */
    void SetupDI( HINSTANCE hInstance, HWND hwndMain );

    /**
     * Polls the keyboard to update the key buffer
     */
    void PollKeyboard();

    /**
     * Checks the key buffer if specified key has been pressed
     *
     * @param keyToCheck specified key to check
     * @pre PollKeyboard has been called to set the key buffer
     * @return true if key is down
     */
    bool IsKeyDown( int keyToCheck );

    /**
     * MousePoll
     *
     *  Polls the mouse to update all the current info about it
     */
    void PollMouse();

    /**
     * IsButtonDown
     *
     * @return true if the passed button is down
     * @pre PollMouse was called to poll the mouse for input
     */
    bool IsButtonDown( int Button );

    /**
     *  GetMousePt
     *
     * @return the mouses X,Y position
     * @pre PollMouse was called
     */
    POINT GetScreenPt();

    /**
     * GetRelativePt
     *
     * @return relative point of x, y movement
     * @pre PollMouse was called
     */
    POINT GetRelativePt();

    /**
     * Currently the mouse window pt is not set, so this function gets garbage
     */
    POINT GameDXInput::GetWindowPt();

    /**
     * Resets Position to 0,0
     */
    void ResetPos();

    /**
     * Centers position to center of window
     */
    void CenterPos();

    /**
     * Creates mouse surface
     *
     * Parameters:
     * pd3dDevice - Initialized Display Device
     */
    void CreateMouseSurf( LPDIRECT3DDEVICE8 m_pd3dDevice, char* pathName );
    
    /**
     * Turn on/off show cursor
     */
    void ShowCursor( bool bShow, LPDIRECT3DDEVICE8 device );

    /**
     * Event called when device is lost.
     */
    void OnLostDevice();

    /**
     * Event called when device is reset.
     */
    void OnResetDevice();


private:
    /**
     * DirectInput device
     */
    LPDIRECTINPUT8 m_pDI;

    /**
     * DirectInput keyboard
     */
    LPDIRECTINPUTDEVICE8 m_pDIKeyboard;

    /**
     * DirectInput mouse
     */
    LPDIRECTINPUTDEVICE8 m_pMouseDev;
    
    /**
     * If devices are initialized
     */
    bool initialized;
    
    /**
     * Polling mouse data
     */
    DIMOUSESTATE m_MouseData;
    
    /**
     * If last mouse poll succeeded
     */
    bool m_bMousePollSuccess;
    
    /**
     * A key buffer
     */
    char m_KeyBuffer[KEYBUFFERSIZE];
    
    /**
     * If last keyboard poll succeeded
     */
    bool m_bKeyPollSuccess;

    /**
     * Mouse cursor surface
     */
    LPDIRECT3DSURFACE8 m_pCursorSurf;
};

#endif