#ifndef KEYFRAMES_H_KMD123
#define KEYFRAMES_H_KMD123

#include "global.h"

/**
 * The base Key class.  Keys can be compared by time.  This class is meant to
 * serve as the base class for specific key types.
 */
class Key {
public:
    Key() {}
    Key( float time ) : time( time ) {}

    float time;

    bool operator < ( const Key& rhs ) const;

    bool operator > ( const Key& rhs ) const;

    bool operator < ( float rhs ) const;

    bool operator > ( float rhs ) const;

    bool operator >= ( float rhs ) const;
    
    bool operator <= ( float rhs ) const;
};

/**
 * A key representing a given rotation at a given time.
 */
class RotKey : public Key {
public:
    RotKey() {}
    RotKey( float time, const Quaternion& rot ) : Key( time ), rot(rot) {}

    Quaternion rot;

    /**
     * Performs spherical linear interpolation between the two keys.
     */
    RotKey interp( const RotKey& next, float atTime ) const;
};

/**
 * A key representing a given vector at a given time.
 */
class VectKey : public Key {
public:
    VectKey() {}
    VectKey( float time, const Vector& v ) : Key( time ), v(v) {}

    Vector v;

    /**
     * Performs linear interpolation between the two keys.
     */
    VectKey interp( const VectKey& next, float atTime ) const;
};

typedef std::vector<RotKey> RKeyList;
typedef RKeyList::iterator RKeyListIter;
typedef RKeyList::const_iterator RKeyListCIter;

typedef std::vector<VectKey> SKeyList;
typedef SKeyList::iterator SKeyListIter;
typedef SKeyList::const_iterator SKeyListCIter;

typedef std::vector<VectKey> TKeyList;
typedef TKeyList::iterator TKeyListIter;
typedef TKeyList::const_iterator TKeyListCIter;

/**
 * A structure of all three types of a keys for a joint.  Contains the joint's
 * name and all of the keyframes for that joint.  Used as a storage area for
 * the parsing code until it breaks the information down and ties it to the
 * Bone.
 */
struct AnimationKeys {
    std::string boneName;
    RKeyList rKeys;
    SKeyList sKeys;
    TKeyList tKeys;
};

typedef std::vector<AnimationKeys> AKeysList;
typedef AKeysList::iterator AKeysListIter;
typedef AKeysList::const_iterator AKeysListCIter;

#endif

