#include "global.h"
#include "SkeletalAnimation.h"
#include "BoneInterpolator.h"
#include "Keyframes.h"
#include "Bone.h"
#include "GameException.h"

using namespace std;

SkeletalAnimation::SkeletalAnimation(
  const AKeysList& keys, const BoneNameSet& boneSet, Bone& rootBone,
  float startTime, float endTime, const std::string& animName)
  : name( animName ) {
    try {
        //Loop over all animation keys to find those only for our bones
        AKeysListCIter iter = keys.begin();
        for ( ; iter != keys.end(); ++iter ) {
            const AnimationKeys& curr = (*iter);
            if ( boneSet.find( curr.boneName ) != boneSet.end() ) {
                //This is a bone in our animation
                Bone* targetBone = rootBone.findBone( curr.boneName );
                if ( !targetBone )
                    throw GameException( "Animation data for " + animName + " contains bone " + curr.boneName + " that is not part of the given Skeleton." );

                BoneInterpolator* nextBi = new BoneInterpolator( targetBone, curr, startTime, endTime );
                bis.push_back( nextBi );
                if ( !nextBi->hasAnyKeys() )
                    throw GameException( "There are no keyframes for bone " + curr.boneName + " during animation " + animName );
            }
        }

        //We should have gotten at least one BoneInterpolator
        if ( bis.empty() )
            throw GameException( "No information was found in the animation keys for animation " + animName );

        assert( bis.size() <= boneSet.size() );

        //Now do the timeshifting
        std::vector<float> fList;

        //Shift by the start time
        BIListIter biIter;
        for ( biIter = bis.begin() ; biIter != bis.end(); ++biIter )
            fList.push_back( (*biIter)->getStartTime() );

        float shiftVal = *(min_element( fList.begin(), fList.end() ));
        for ( biIter = bis.begin() ; biIter != bis.end(); ++biIter )
            (*biIter)->timeShift( shiftVal );

        //Find the anim length
        fList.clear();
        for ( biIter = bis.begin() ; biIter != bis.end(); ++biIter )
            fList.push_back( (*biIter)->getEndTime() );

        length = *(max_element( fList.begin(), fList.end() ) );

    } catch ( ... ) {
        cleanUp();
        throw;
    }
}

SkeletalAnimation::~SkeletalAnimation() {
    cleanUp();
}

void SkeletalAnimation::cleanUp() {
    BIListIter iter = bis.begin();
    for ( ; iter != bis.end(); ++iter )
        delete (*iter);
}
    
float SkeletalAnimation::getLength() const {
    return length;
}

const std::string& SkeletalAnimation::getName() const {
    return name;
}

void SkeletalAnimation::animate( float t ) {
    BIListIter iter = bis.begin();
    for ( ; iter != bis.end(); ++iter )
        (*iter)->animate( t );
}
