#ifndef SKELETALANIMATION_H_EJR421
#define SKELETALANIMATION_H_EJR421

#include "global.h"
#include "Keyframes.h"
class BoneInterpolator;
class Bone;

/**
 * The SkeletalAnimation represents a atomic animation component. It contains
 * a list of BoneInterpolator objects under a common name and an animation
 * length.  An example of a single SkeletalAnimation would be a walk cycle
 * that operates only on the leg and hip joints.
 */
class SkeletalAnimation {
public:
    /**
     * Constructs a new SkeletalAnimation object, given the data in the X file
     * for an AnimationSet block and the ranges to pull out data from.  Time
     * values are given in seconds.
     *
     * @param keys Data object to the AnimationSet block
     * @param boneSet Set of bones to filter (only looks at bones in this
     *                list).
     * @param rootBone The root bone of the skeleton to perform operations on.
     * @param startTime Starting time to look for animation keys.
     * @param endTime Ending time to look for animation keys.
     * @param animName the name assigned to this set of animation.
     */
    SkeletalAnimation( const AKeysList& keys, const BoneNameSet& boneSet,
                       Bone& rootBone, float startTime, float endTime,
                       const std::string& animName);

    /**
     * Destructor.
     */
    ~SkeletalAnimation();
    
    /**
     * Returns the length of this animation, in seconds.
     */
    float getLength() const;

    /**
     * Returns the name of this SkeletalAnimation object.
     */
    const std::string& getName() const;

    /**
     * Animates the skeleton at the given time t.  If looping is to performed,
     * the caller needs to implement the wrapping of t.  The start of the
     * animation is at 0 sections, and the end is determined by the getLength
     * method.
     *
     * If t is less than 0, then the animation will be based on the first key
     * for every BoneInterpolator, with no interpolation.  Similarly for (t >
     * getLength()), the last key will be used.
     *
     * @see BoneInterpolator::animate
     */
    void animate( float t );
    
private:
    /**
     * Performs cleanUp (destructor-like) operations.
     */
    void cleanUp();

    SkeletalAnimation();
    SkeletalAnimation( const SkeletalAnimation& o );
    SkeletalAnimation& operator = ( const SkeletalAnimation& rhs );
    
    std::string name;
    float length;

    typedef std::vector<BoneInterpolator*> BIList;
    typedef BIList::iterator BIListIter;
    typedef BIList::const_iterator BIListCIter;

    BIList bis;
};

#endif

