#include "global.h"
#include "SkinnedModel.h"
#include "GameException.h"

#include <rmxftmpl.h> //X File D3DRM templates

using namespace std;

SkinnedModel::SkinnedModel()
: mesh(NULL), boneCombinationBuf(NULL), maxInfl(0), numMats(0),
  materials(NULL), textures(NULL), rootBone(NULL), boneNamesBuf(NULL),
  boneOffMats(NULL) {
}

SkinnedModel::SkinnedModel( const std::string& fileName )
: mesh(NULL), boneCombinationBuf(NULL), maxInfl(0), numMats(0),
  materials(NULL), textures(NULL), rootBone(NULL), boneNamesBuf(NULL),
  boneOffMats(NULL) {
    HRESULT r = 0;
    LPDIRECTXFILE xFile = NULL;
    LPDIRECTXFILEENUMOBJECT xEnum = NULL;
    LPDIRECTXFILEDATA xObj = NULL;

    try {
        //File interface create and open with enum object.
        r = DirectXFileCreate( &xFile );
        if ( FAILED( r ) ) throw GameException( "DirectXFileCreate Failure", r );

        r = xFile->RegisterTemplates((LPVOID)D3DRM_XTEMPLATES, D3DRM_XTEMPLATE_BYTES);
        if ( FAILED( r ) ) throw GameException( "XFile RegisterTemplates Failure", r );

        r = xFile->CreateEnumObject( (LPVOID)fileName.c_str(), DXFILELOAD_FROMFILE, &xEnum);
        if ( FAILED( r ) ) throw GameException( "XFile CreateEnumObject Failure", r );

        //Load bones/mesh
        r = xEnum->GetNextDataObject( &xObj );
        if ( FAILED( r ) ) throw GameException( "Error finding skeletal information", r );
        loadSkeleton( xObj );
        SAFERELEASE( xObj );

    } catch ( GameException& e ) {
        //Cleanup
        SAFERELEASE( xFile );
        SAFERELEASE( xEnum );
        SAFERELEASE( xObj );
        cleanUp();
        e.addMessage( "For SkinnedModel \"" + fileName + "\"" );
        throw e;
    }

    //Cleanup
    SAFERELEASE( xFile );
    SAFERELEASE( xEnum );
    SAFERELEASE( xObj );
}

std::string SkinnedModel::getDXDataObjectName( LPDIRECTXFILEDATA xObj ) {
    HRESULT r = 0;
    DWORD nameLen;

    r = xObj->GetName( NULL, &nameLen );
    if ( FAILED( r ) ) throw GameException( "Error trying to get name from X file object", r );

    if ( nameLen > 0 ) {
        char * nameStr = new char[ nameLen ];

        r = xObj->GetName( nameStr, &nameLen );
        if ( FAILED( r ) ) throw GameException( "Error trying to get name from X file object", r );

        std::string ret( nameStr );
        delete[] nameStr;
        return ret;

    } else
        return std::string( "" );
}

void SkinnedModel::loadSkeleton( LPDIRECTXFILEDATA xObj ) {
    const GUID* oType;
    if ( FAILED( xObj->GetType( &oType ) ) || *oType != TID_D3DRMFrame )
        throw GameException( "XFile Parser: The first data in the file must be the root frame." );

    //Load the root bone
    rootBone = new Bone();
    rootBone->setName( getDXDataObjectName( xObj ) );

    loadFrame( xObj, rootBone );

    //The skeleton is completely loaded here, so we resolve the bones
    resolveBones();
}

void SkinnedModel::loadFrame( LPDIRECTXFILEDATA xObj, Bone* currBone ) {
    LPDIRECTXFILEDATA xChildDat;
    LPDIRECTXFILEOBJECT xChildObj;

    while ( SUCCEEDED( xObj->GetNextObject( &xChildObj ) ) ) {
        if ( SUCCEEDED( xChildObj->QueryInterface( IID_IDirectXFileData, (LPVOID*)&xChildDat ) ) ) {
            processFrame( xChildDat, currBone );
            xChildDat->Release();
        } else {
            xChildObj->Release();
            throw GameException( "XFile Parser: Unrecognized element in XFile." );
        }
        xChildObj->Release();
    }
}

void SkinnedModel::processFrame( LPDIRECTXFILEDATA xObj, Bone* currBone ) {
    const GUID* oType = NULL;
    HRESULT r = 0;

    r = xObj->GetType( &oType );
    if ( FAILED( r ) ) throw GameException( "XFile Parser: Could not read object's type.", r );

    if ( *oType == TID_D3DRMFrameTransformMatrix ) {
        D3DXMATRIX* data;
        DWORD size;

        r = xObj->GetData( NULL, &size, (PVOID*)&data );
        if ( FAILED( r ) ) throw GameException( "XFile Parser: Could not pull data out of object.", r );
        currBone->setReferenceMatrix( *data );
        currBone->setTransform( *data );

    } else if ( *oType == TID_D3DRMMesh ) {
        if ( mesh )
            throw GameException( "This XFile Parser only supports 1 mesh per model." );
        
        loadMesh( xObj );

    } else if ( *oType == TID_D3DRMFrame ) {
        //Create a new Bone
        Bone* newBone = new Bone();
        newBone->setName( getDXDataObjectName( xObj ) );
        currBone->addChild( newBone );
        loadFrame( xObj, newBone );

    } else {
        throw GameException( "XFile Parser: Unrecognized element in XFile(2)." );
    }
}

void SkinnedModel::loadMesh( LPDIRECTXFILEDATA xObj ) {
    HRESULT r = 0;
    LPD3DXBUFFER matBuf = NULL; //materials for the mesh
    LPD3DXBUFFER adjBuf = NULL; //vertex adjencency info
    LPD3DXSKINMESH skinMesh = NULL;
    DWORD * meshAdjInfo = NULL; //adjacency array

    r = D3DXLoadSkinMeshFromXof( xObj, D3DXMESH_MANAGED, g_dx->getDeviceHandle(),
        &adjBuf, &matBuf, (DWORD*)&numMats, &boneNamesBuf, &boneOffMats, &skinMesh );
    if ( FAILED( r ) ) throw GameException( "XFile Parser: Error loading skin mesh.", r );

    numBones = skinMesh->GetNumBones();
    if ( numBones <= 0 ) throw GameException( "XFile Parser: No bones found in skin mesh." );

    //Load in the skinning data.
    //Convert to a blended mesh
    r = skinMesh->ConvertToBlendedMesh( D3DXMESH_WRITEONLY | D3DXMESH_MANAGED, (LPDWORD)adjBuf->GetBufferPointer(),
        NULL, &numComb, &boneCombinationBuf, NULL, NULL, &mesh );
    if ( FAILED( r ) ) throw GameException( "XFile Parser: Could not convert skin mesh to blended mesh.", r );

    //Get how many bones we render per call.
    if( (mesh->GetFVF() & D3DFVF_POSITION_MASK) != D3DFVF_XYZ )
        maxInfl = 1 + ((mesh->GetFVF() & D3DFVF_POSITION_MASK) - D3DFVF_XYZRHW) / 2;
    else
        maxInfl = 1;

    //The mesh is loaded, let's load the materials:
    materials = new D3DMATERIAL8[ numMats ];
    textures = new Texture[ numMats ];
    int i = 0;
    for ( i = 0; i < numMats; ++i )
        textures[i] = NULL;

    D3DXMATERIAL* modelMats = (D3DXMATERIAL*)matBuf->GetBufferPointer();

    for ( i = 0; i < numMats; ++i ) {
        materials[ i ] = modelMats[ i ].MatD3D;
        materials[ i ].Ambient = materials[ i ].Diffuse;

        //MAYA HACK
        materials[ i ].Power *= 100.0f;

        if ( modelMats[i].pTextureFilename ) {
            //Attempt to load this texture
            //Extract just the filename part.
            string tfileName( modelMats[i].pTextureFilename );
            string::size_type pos = tfileName.find_last_of( "\\/" );
            if ( pos != string::npos ) {
                tfileName = string( tfileName, pos+1 );
            }

            //Add the directory
            tfileName = "models/" + tfileName;

            //Attempt to load
            r = D3DXCreateTextureFromFile(
                g_dx->getDeviceHandle(),
                tfileName.c_str(),
                &textures[ i ] );

            if ( FAILED( r ) )
                throw GameException( "Could not load texture " + tfileName + ".", r );

        } else {
            textures[i] = NULL;
        }
    }
}

/**
 * After the model is loaded, we need to resolve the bone names referenced in
 * the mesh to the bones loaded from the skelton, and preserve their order for
 * use in rendering.  The boneList serves this purpose.  We also take this
 * opportunity to set the offset matrix for the bone.
 */
void SkinnedModel::resolveBones() {
    char** boneNames = (char**)boneNamesBuf->GetBufferPointer();
    Matrix* offMats = (Matrix*)boneOffMats->GetBufferPointer();

    for ( int c = 0; c < numBones; ++c ) {
        Bone* currBone = rootBone->findBone( boneNames[c] );
        if ( !currBone) throw GameException( "XFile Parser: Mesh referenced bone '" + string(boneNames[c]) + "' that was not found in the skeleton." );

        currBone->setOffsetMatrix( offMats[c] );
        boneList.push_back( currBone );
    }

    SAFERELEASE( boneNamesBuf );
    SAFERELEASE( boneOffMats );
}

void SkinnedModel::cleanUp() {
    delete rootBone;
    rootBone = NULL;

    SAFERELEASE( mesh );
    SAFERELEASE( boneCombinationBuf );

    delete[] materials;
    materials = NULL;

    if ( textures ) {
        for ( int i = 0; i < numMats; ++i ) {
            if ( textures[ i ] )
                textures[ i ]->Release();
        }

        delete[] textures;
        textures = NULL;
    }

    //these should normally be NULL on normal destruction.
    SAFERELEASE( boneNamesBuf );
    SAFERELEASE( boneOffMats );
}

SkinnedModel::~SkinnedModel() {
    cleanUp();
}

Bone* SkinnedModel::getSkeleton() {
    return rootBone;
}

void SkinnedModel::setToReference() {
    BoneListIter iter = boneList.begin();
    for ( ; iter != boneList.end(); ++iter ) {
        (*iter)->setTransform( (*iter)->getReferenceMatrix() );
    }
}

void SkinnedModel::render() {
    Matrix m;
    D3DXMatrixIdentity( &m );
    render( m );
}

void SkinnedModel::render( const Matrix& worldMat ) {
    rootBone->updateHierarchy( worldMat );

    LPD3DXBONECOMBINATION boneCombonations =
        (LPD3DXBONECOMBINATION)boneCombinationBuf->GetBufferPointer();

    GraphicsDevice d = g_dx->getDeviceHandle();
    HRESULT r = 0;
    Matrix skinMat;

    //Turn on vertex blending
    r = d->SetRenderState( D3DRS_VERTEXBLEND, maxInfl-1 );
    assert( SUCCEEDED( r ) );

    //loop over all of the bone combonations (attrib set)
    for ( DWORD comb = 0; comb < numComb; ++comb ) {
        for ( DWORD b = 0; b < maxInfl; ++b ) {
            //check all bones in this attribute set
            DWORD boneId = boneCombonations[comb].BoneId[b];
            DWORD oldId = (comb == 0) ? -1 : boneCombonations[comb-1].BoneId[b];
            if ( boneId != UINT_MAX && (boneId != oldId ) ) {
                //Only update the bone matrix if it has changed.
                boneList[ boneId ]->loadSkinMatrix( skinMat );
                r = d->SetTransform( D3DTS_WORLDMATRIX(b), &skinMat );
                assert( SUCCEEDED( r ) );
            }
        }

        //Only update material if it is a different material
        DWORD matId = boneCombonations[comb].AttribId;
        DWORD oldMatId = (comb == 0) ? -1 : boneCombonations[comb-1].AttribId;
        if ( matId != oldMatId ) {
            r = d->SetMaterial( &materials[matId] );
            assert( SUCCEEDED( r ) );
            r = d->SetTexture( 0, textures[matId] );
            assert( SUCCEEDED( r ) );
        }

        r = mesh->DrawSubset( comb );
        assert( SUCCEEDED( r ) );
    }

    //Turn off vertex blending
    r = d->SetRenderState( D3DRS_VERTEXBLEND, D3DVBF_DISABLE );
    assert( SUCCEEDED( r ) );
}

