#ifndef SKINNEDMODEL_H_WNR262
#define SKINNEDMODEL_H_WNR262

#include "global.h"
#include "Bone.h"

/**
 * The SkinnedModel class is a model that is defined by a mesh attached to a
 * skeleton. By itself it neither contains nor implements any animation, but
 * the Skeleton can be addressed directly to modify the skeleton, and the mesh
 * will be rendered using the changes made to the Skeleton.
 */
class SkinnedModel {
  public:
    /**
     * Loads a SkinnedModel from a X file with its skeleton.  No animations
     * are loaded.  Only one Mesh is allowed in the X file.  A GameException
     * is thrown if an error occured while loading.
     */
    SkinnedModel( const std::string& fileName );

    /**
     * Destructor.
     */
    virtual ~SkinnedModel();

    /**
     * Returns the root Bone for the Skeleton.  The Skeleton can then be
     * modified or viewed through the returned root, and any changes made will
     * be reflected in the next render call.
     */
    Bone* getSkeleton();

    /**
     * Places all joints in the reference pose.
     */
    void setToReference();
    
    /**
     * Renders the model at the origin of the world.  Same effect as calling
     * render with the identity matrix.  The current world matrix set in the
     * API is overwritten.
     */
    void render();

    /**
     * Renders the model using the given world matrix.  Using this input world
     * matrix, you can rotate, translate, and scale the entire model as a
     * whole.  This is necessary because the hardware matrices must be
     * overwritten due to the skinning process, so the Skeleton must be
     * modified by the given Matrix.
     */
    void render( const Matrix& worldMat );

    /**
     * A utility method that returns the name of a data object in an X file as
     * a std::string object.
     */
    static std::string getDXDataObjectName( LPDIRECTXFILEDATA xObj );

  protected:
    /**
     * The SkinnedModel default constructor is protected because the procedure
     * for loading animations with the model is slightly different.
     */
    SkinnedModel();

    /**
     * This does the real work of loading the SkinnedModel.  It is called by
     * the public constructor, but as the loading process differs slightly in
     * AnimatedModel, AnimatedModel will call this function in its
     * constructor.
     */
    void loadSkeleton( LPDIRECTXFILEDATA xObj );
    
    /**
     * Performs destructor functionality, used if exception is to be thrown
     * from contructor
     */
    virtual void cleanUp();

  private:
    void loadFrame( LPDIRECTXFILEDATA xObj, Bone* currBone );
    void processFrame( LPDIRECTXFILEDATA xObj, Bone* currBone );
    void loadMesh( LPDIRECTXFILEDATA xObj );
    void resolveBones();

    ID3DXMesh* mesh;
    DWORD maxInfl; //max number of bones per vertex
    DWORD numComb; //number of bone combonations used for rendering
    LPD3DXBUFFER boneCombinationBuf; //Bone combinations buffer

    int numMats;
    D3DMATERIAL8* materials;
    Texture* textures;

    Bone* rootBone;
    BoneList boneList; //linear list of bones, that match the bone comb buf

    //temporary info
    LPD3DXBUFFER boneNamesBuf;
    int numBones;
    LPD3DXBUFFER boneOffMats; //offset matrices
};

#endif

