/**
 * CG3 Skeletal Animation Project
 * Jason Winnebeck
 * May 12, 2003
 */

#include "global.h"
#include "AnimatedModel.h"
#include "StaticModel.h"
#include "ConfigFile.h"

const char* WINDOW_TITLE = "CG3: Skeletal Animation Project";

static bool dispFps = true;

GameDX* g_dx = NULL;
GameDXInput* g_DI = NULL;
Window* g_win = NULL;
ConfigFile* g_cfg = NULL;

static bool s_isPaused = false;

using namespace std;

static string inputFileName;
static string rootPath;

void initializeGame( HINSTANCE hInstance, int nShowCmd );
int messageLoop();
void cycleAnim( AnimatedModel& m, int setIdx );
void cycleLoop( AnimatedModel& m, int setIdx );

//This is a quick hack for the keys to cycle thru the animations
const int KEYDOWN_SIZE = 10;
static bool keyDown[KEYDOWN_SIZE];

void trimString( string& str ) {
    str = str.erase( str.find_last_not_of( " \t\n\"" ) + 1 ); //right trim
    str.erase( 0, str.find_first_not_of( " \t\n\"" ) ); //left trim
}

int APIENTRY WinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance,
                      LPSTR lpCmdLine, int nShowCmd) {
    int ret = 0;

    for ( int c = 0; c < KEYDOWN_SIZE; ++c )
        keyDown[c] = false;

    TCHAR buf[512];
    GetModuleFileName( GetModuleHandle( NULL ), buf, 512 );
    rootPath = buf;
    //remove the executable's name, leave trailing slash.
    rootPath.erase( rootPath.find_last_of( '\\' )+1 );

    try {
        initializeGame( hInstance, nShowCmd );
        ret = messageLoop();

    } catch ( const GameException& e ) {
        e.showMessageBox();
    }

    delete g_DI;
    delete g_dx;
    delete g_win;

    return ret;
}

void initializeGame( HINSTANCE hInstance, int nShowCmd ) {
    //Load the configuration file
    g_cfg = new ConfigFile( rootPath + "config.ini" );
    int fullscreen;
    g_cfg->getVar( "Preferences", "fullscreen", fullscreen, 0 );

    StaticModel::setModelDirectory( rootPath + "models/" );

    // Setup Window
    int width, height;
    g_cfg->getVar( "Preferences", "resx", width, 800 );
    g_cfg->getVar( "Preferences", "resy", height, 600 );

    g_win = new Window;
    g_win->setWindowed( (fullscreen) ? FALSE : TRUE );
    g_win->registerClass( hInstance, WINDOW_TITLE );
    try {
        g_win->initInstance( hInstance, nShowCmd, WINDOW_TITLE, WINDOW_TITLE, width, height );
    } catch ( const GameException& ) {
        //I do the delete here, because GameException uses g_win to generate the
        //MessageBox, so g_win has to have a valid HWND.
        delete g_win;
        g_win = NULL;
        throw;
    }

    // Setup DirectX and tie it to our window
    g_dx = new GameDX;
    g_dx->setupD3D();

    // Initialize DirectInput
    g_DI = new GameDXInput();
    g_DI->SetupDI( hInstance, g_win->getHwnd() );

    //Set view matrix
    float posX, posY, posZ, laX, laY, laZ;
    g_cfg->getVar( "CameraPos", "x", posX, 2.0f );
    g_cfg->getVar( "CameraPos", "y", posY, 0.75f );
    g_cfg->getVar( "CameraPos", "z", posZ, 0.0f );
    g_cfg->getVar( "CameraLookAt", "x", laX, 0.0f );
    g_cfg->getVar( "CameraLookAt", "y", laY, 0.75f );
    g_cfg->getVar( "CameraLookAt", "z", laZ, 0.0f );

    D3DXMATRIX view;
    D3DXMatrixLookAtLH(
        &view,
        &D3DXVECTOR3( posX, posY, posZ ),
        &D3DXVECTOR3( laX, laY, laZ ),
        &D3DXVECTOR3( 0.0f, 1.0f, 0.0f ) );
    g_dx->SetTransform( D3DTS_VIEW, &view );

    //Set projection matrix
    float nearZ, farZ;
    g_cfg->getVar( "Fustrum", "near", nearZ, 1.0f );
    g_cfg->getVar( "Fustrum", "far", farZ, 20.0f );

    float aspect = (float)g_win->getWidth() / (float)g_win->getHeight();
    D3DXMATRIX projection;
    D3DXMatrixPerspectiveFovLH( &projection, D3DX_PI/4.0f, aspect, nearZ, farZ );
    g_dx->SetTransform( D3DTS_PROJECTION, &projection );

    //Setup a light for this scene
    D3DLIGHT8 light;
    light.Type = D3DLIGHT_DIRECTIONAL;
    light.Direction = Vector( -0.5f, -1.0f, 0.0f );
    //light.Type = D3DLIGHT_POINT;
    light.Diffuse = D3DXCOLOR( 0.6f, 0.6f, 0.6f, 1.0f );
    light.Specular = D3DXCOLOR( 1.0f, 1.0f, 1.0f, 1.0f );
    light.Position = D3DXVECTOR3( 0.0f, 1.0f, 0.0f );
    light.Ambient = D3DXCOLOR( 0.1f, 0.1f, 0.1f, 1.0f );
    light.Range = 10.0f;
    light.Attenuation0 = 0.0f;
    light.Attenuation1 = 0.0f;
    light.Attenuation2 = 1.0f/0.8f;
    g_dx->LightEnable( 0, TRUE ); //use default directional light pointing towards +z
    g_dx->SetLight( 0, &light );

    D3DLIGHT8 light2 = light;
    light2.Direction = Vector( 0.0f, 0.3f, 1.0f );
    light2.Diffuse = D3DXCOLOR( 0.6f, 0.6f, 0.6f, 1.0f );
    light2.Ambient = D3DXCOLOR( 0.0f, 0.0f, 0.0f, 0.0f );
    g_dx->SetLight( 1, &light2 );
    g_dx->LightEnable( 1, TRUE );

    g_dx->SetClearColor( D3DCOLOR_XRGB( 16, 32, 48 ) );
}

int messageLoop() {
    GraphicsDevice dev = g_dx->getDeviceHandle();

    float dt;
    float totalTime( 0.0f );
    int slowmo = 0;
    g_cfg->getVar( "Preferences", "slowmo", slowmo, slowmo );

    string modelName;
    //pick an unlikely nodel name as default to check for unspecified
    g_cfg->getVar( "Model", "modelName", modelName, "9ck24md" );
    if ( modelName == "9ck24md" )
        throw GameException( "Must specify model name in config.ini." );
    trimString( modelName );
    AnimatedModel m( modelName );

    //Perform coordinate system fix for Maya
    int flipZ;
    g_cfg->getVar( "Model", "RHtoLH", flipZ, 1 );
    DWORD cullMode = D3DCULL_CCW;

    if ( flipZ ) {
        //HACK HACK fix this
        Bone* b = m.getSkeleton()->findBone( "SCENE_ROOT" );
        Matrix mat;
        D3DXMatrixScaling( &mat, 1.0f, 1.0f, -1.0f );
        b->setTransform( mat );
        cullMode = D3DCULL_CW;
    }

    //Output skeleton for debugging.
    string skelDebugFileName;
    g_cfg->getVar( "Skeleton", "outputFile", skelDebugFileName, "skeleton.txt" );
    trimString( skelDebugFileName );
    ofstream skelDebug( skelDebugFileName.c_str() );
    m.getSkeleton()->printHierarchy( skelDebug );
    skelDebug.close();

    //rotation controls
    float rotY = 0.0f;

    // Intialize timer
    Timer timer;
    timer.setStartTime();

    MSG msg;
    while(TRUE) {
        if ( !s_isPaused ) {
            //check and see if there is a message
            if (PeekMessage( &msg, NULL, 0, 0, PM_REMOVE)) {
                //check and see if this is the quit message
                if (msg.message == WM_QUIT)
                    //exit the message loop
                    break;
                //else do what we would have before, namely translate and dispatch
                TranslateMessage( &msg );
                DispatchMessage( &msg );
            }
            else {
                HRESULT r = 0;
                
                //Timing
                dt = timer.getFrameDt();
                if ( slowmo )
                    dt /= 5.0f;
                totalTime += dt;
                
                //Check for keyboard
                g_DI->PollKeyboard();

                if( g_DI->IsKeyDown( DIK_F1 ) )
                    g_win->toggleWindowed();
                if ( g_DI->IsKeyDown( DIK_ESCAPE ) )
                    PostQuitMessage( 0 );
                if ( g_DI->IsKeyDown( DIK_S ) )
                    slowmo = 1;
                if ( g_DI->IsKeyDown( DIK_N ) )
                    slowmo = 0;
                if ( g_DI->IsKeyDown( DIK_R ) )
                    m.setToReference();

                if ( g_DI->IsKeyDown( DIK_LEFT ) )
                    rotY -= dt * 2.0f;
                else if ( g_DI->IsKeyDown( DIK_RIGHT ) )
                    rotY += dt * 2.0f;

                //This is a hack, someday replace it with real code
                //(it assumes that the number key codes are in order.)
                for ( int currKey = 0; currKey < KEYDOWN_SIZE; ++currKey ) {
                    if ( g_DI->IsKeyDown( DIK_1 + currKey ) ) {
                        if ( !keyDown[currKey] ) {
                            keyDown[currKey] = true;
                            if ( g_DI->IsKeyDown( DIK_L ) )
                                cycleLoop( m, currKey );
                            else
                                cycleAnim( m, currKey );
                        }
                    } else
                        keyDown[currKey] = false;
                }

                m.update( dt );

                //Draw the next frame
                r = dev->BeginScene();
                if ( SUCCEEDED( r ) ) {
                    g_dx->Clear();

                    dev->SetRenderState( D3DRS_CULLMODE, cullMode );

                    Matrix worldMat;
                    D3DXMatrixRotationYawPitchRoll( &worldMat, rotY, 0.0f, 0.0f );
                    m.render( worldMat );

                    char buf[128];
                    RECT fontRect = {8, 8, 200, 200};

                    // Display frame rate, current time, and instructions
                    if( dispFps ) {
                        sprintf( buf, "FPS: %i", timer.getFrameRate() );
                        g_dx->getFont()->DrawText( buf, -1, &fontRect, DT_NOCLIP, D3DCOLOR_XRGB(255,255,255) );

                        fontRect.top = 28;
                        sprintf( buf, "Time (s): %f", totalTime );
                        g_dx->getFont()->DrawText( buf, -1, &fontRect, DT_NOCLIP, D3DCOLOR_XRGB(255,255,255) );

                        fontRect.top = g_win->getHeight() - 44;
                        sprintf( buf, "Number keys = cycle animation, Number + L = toggle looping, R = Reference pose" );
                        g_dx->getFont()->DrawText( buf, -1, &fontRect, DT_NOCLIP, D3DCOLOR_XRGB(255,255,255) );

                        fontRect.top = g_win->getHeight() - 24;
                        sprintf( buf, "S = SloMo, N = normal speed, F1 = fullscreen toggle, arrows = rotate, ESC = quit" );
                        g_dx->getFont()->DrawText( buf, -1, &fontRect, DT_NOCLIP, D3DCOLOR_XRGB(255,255,255) );
                    }

                    //Print anim sets and currently running anim
                    fontRect.top = 68;
                    for ( int s = 0; s < m.getNumSets(); ++s ) {
                        const string& setName = m.getSetName( s );
                        int a = m.getActiveAnim( s );
                        const string& animName = m.getAnimName( s, a );
                        string loopingString = (m.isLooping(s)) ? " (looping)" : "";

                        sprintf( buf, "%d: %s: (%d)%s%s", s+1, setName.c_str(), a, animName.c_str(), loopingString.c_str() );
                        g_dx->getFont()->DrawText( buf, -1, &fontRect, DT_NOCLIP, D3DCOLOR_XRGB(255,255,255) );
                        fontRect.top += 20;
                    }

                    dev->EndScene();
                    dev->Present( NULL, NULL, NULL, NULL );

                    timer.calcFrameCount();
                } else {
                    TRACE( "BeginScene call failed!" );
                }
            }

        } else {
            //We want to wait for messages when paused.
            GetMessage( &msg, NULL, 0, 0 );
            //check and see if this is the quit message
            if (msg.message == WM_QUIT) {
                //exit the message loop
                break;
            }
            //else do what we would have before, namely translate and dispatch
            TranslateMessage( &msg );
            DispatchMessage( &msg );
        }
    }

    return (int)msg.wParam;
}

void cycleAnim( AnimatedModel& m, int setIdx ) {
    if ( setIdx >= 0 && setIdx < m.getNumSets() ) {
        int a = m.getActiveAnim( setIdx ) + 1;
        if ( a == m.getNumAnims( setIdx ) )
            a = -1;
        m.setActiveAnim( setIdx, a );
    }
}

void cycleLoop( AnimatedModel& m, int setIdx ) {
    if ( setIdx >= 0 && setIdx < m.getNumSets() ) {
        m.setLooping( setIdx, !m.isLooping( setIdx ) );
    }
}

LRESULT CALLBACK GlobalWndProc( HWND hWnd, UINT message, 
                                WPARAM wParam, LPARAM lParam ) {
    HDC hdc;
    PAINTSTRUCT ps; //structure used by Windows to paint things
    switch(message) {

        case WM_CREATE:
            return DefWindowProc( hWnd, message, wParam, lParam );
            break;
        case WM_PAINT:
            hdc = BeginPaint(hWnd, &ps);
            //do any windows drawing here
            EndPaint(hWnd, &ps);
            break;
        case WM_CLOSE:
            TRACE("Posting Destroy");
            PostQuitMessage(0);
            break;
        case WM_SIZE:
            if ( wParam == SIZE_RESTORED || wParam == SIZE_MAXIMIZED ) {
                s_isPaused = false;
            } else if ( wParam == SIZE_MINIMIZED ) {
                s_isPaused = true;
            }
            break;
        case WM_ACTIVATE:
            if( (LOWORD(wParam) != WA_INACTIVE) && (HIWORD(wParam) == 0) ) {
                // If window loses focus, need to reinitialize keyboard
                if( g_DI != NULL ) {
                    g_DI->SetupDI( g_win->get_hInstance(), g_win->getHwnd() );
                }
            }
            break;
        case WM_SETCURSOR:
            if ( g_win->getWindowed() == FALSE ) {
                SetCursor( NULL );
                if( g_dx->getDeviceHandle() != NULL ) {
                    g_dx->getDeviceHandle()->ShowCursor( FALSE );
                }
            }
            else {
                return DefWindowProc( hWnd, message, wParam, lParam );
            }
            break;
        default:
            return DefWindowProc(hWnd, message,wParam,lParam);
    }

    return 0;
}

void doTrace( const char* fn, int lineno, const char* msg, ... ) {
    va_list arg;
    va_start( arg, msg );
    char buf[ 1024 ];
    vsprintf( buf, msg, arg );
    va_end( arg );

    char buf2[ 1024 ];
    sprintf( buf2, "File: %s, Line: %i: %s\n", fn, lineno, buf );

    OutputDebugString( buf2 );
}
